# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import gc
import os
import re
import threading
from abc import ABC, abstractmethod
from collections import defaultdict

from msprobe.core.common.exceptions import DistributedNotInitializedError
from msprobe.core.common.runtime import Runtime
from msprobe.core.common.utils import Const, ThreadSafe
from msprobe.core.dump.data_dump.data_processor.base import (ModuleBackwardInputsOutputs, ModuleForwardInputsOutputs)


class HookSet:
    def __init__(
            self,
            forward_pre_hook=None,
            forward_hook=None,
            backward_pre_hook=None,
            backward_hook=None,
            distributed_forward_hook=None
    ):
        self.forward_pre_hook = forward_pre_hook
        self.forward_hook = forward_hook
        self.backward_pre_hook = backward_pre_hook
        self.backward_hook = backward_hook
        self.distributed_forward_hook = distributed_forward_hook


class BaseHookManager(ABC):
    inner_switch = defaultdict(bool)
    inner_api_count = defaultdict(int)
    hook_handle_dict = {}
    params_grad_info = {}
    grad_hook_call = {}

    def __init__(self, data_collector, config):
        self.data_collector = data_collector
        self.config = config
        self.current_pid = self._pid
        self.logger = None
        self._init_specific_components()

    @property
    def _pid(self):
        return os.getpid()

    @staticmethod
    def reset_status():
        BaseHookManager.inner_switch = defaultdict(bool)
        BaseHookManager.inner_api_count = defaultdict(int)
        BaseHookManager.params_grad_info.clear()

    @staticmethod
    def ensure_gc_enabled():
        is_gc_disabled = not gc.isenabled()
        if is_gc_disabled:
            gc.enable()
        return is_gc_disabled

    @staticmethod
    def restore_gc_state(original_state):
        if original_state:
            gc.disable()

    @staticmethod
    def _clear_input_kwargs(module, tid):
        if hasattr(module, 'msprobe_input_kwargs') and tid in module.msprobe_input_kwargs:
            del module.msprobe_input_kwargs[tid]

    @staticmethod
    def _get_grad_hook_call_index(ori_name, param_name):
        if ori_name not in BaseHookManager.grad_hook_call:
            BaseHookManager.grad_hook_call[ori_name] = [0, param_name]
        else:
            if BaseHookManager.grad_hook_call.get(ori_name)[1] == param_name:
                BaseHookManager.grad_hook_call[ori_name][0] += 1
        return BaseHookManager.grad_hook_call.get(ori_name)[0]

    @staticmethod
    @abstractmethod
    def _no_grad_context():
        pass

    @staticmethod
    @abstractmethod
    def _add_count(name):
        pass

    @staticmethod
    @abstractmethod
    def _get_count(name):
        pass

    @staticmethod
    @abstractmethod
    def _process_kwargs_and_output(module, tid, hook_type, kwargs_or_output, output_or_kwargs):
        pass

    @staticmethod
    @abstractmethod
    def _get_current_rank():
        pass

    @abstractmethod
    def build_hook(self, hook_type, name):
        pass

    @abstractmethod
    def _register_forward_hook(self, module, api_name):
        pass

    @abstractmethod
    def _register_backward_hook(self, module, full_backward_name, args):
        pass

    @abstractmethod
    def _register_backward_pre_hook(self, module, full_backward_name, args, kwargs, output):
        pass

    @abstractmethod
    def _get_params_dict(self, module):
        pass

    @abstractmethod
    def _need_exchange(self, module):
        pass

    @abstractmethod
    def _register_param_hook(self, name, module, params_dict):
        pass

    @abstractmethod
    def _init_specific_components(self):
        """初始化框架特定组件"""
        pass

    def is_child_process(self):
        return self.current_pid != self._pid

    def _maybe_update_dump_dir(self):
        if not self.data_collector.dump_file_path or Runtime.current_rank is not None:
            return

        parent_dir = os.path.dirname(self.data_collector.dump_file_path)
        is_proc_dir = bool(re.search(r'proc\d+$', parent_dir))
        current_rank = None
        try:
            current_rank = self._get_current_rank()
        except DistributedNotInitializedError:
            pass
        else:
            Runtime.current_rank = current_rank
        if is_proc_dir and current_rank is not None:
            new_rank_dir = os.path.join(os.path.dirname(parent_dir), f"{Const.RANK}{current_rank}")
            os.rename(parent_dir, new_rank_dir)
            self.data_collector.replace_proc_with_rank(new_rank_dir)
            self.logger.info(f"Successfully replaced proc path<{parent_dir}> with rank path<{new_rank_dir}>")

    def _should_execute_hook(self, hook_type, tid, is_forward=True):
        is_api_hook = hook_type == Const.API
        if self.is_child_process():
            return False
        if BaseHookManager.inner_switch[tid]:
            return False
        if not is_api_hook and not Runtime.is_running:
            return False
        elif is_api_hook and is_forward and not Runtime.is_running:
            return False
        if not self.data_collector or self.data_collector.data_processor.is_terminated:
            return False
        return True

    def _build_forward_pre_hook(self, hook_type, api_name):
        def forward_pre_hook(module, args, kwargs=None):
            if hook_type == Const.MODULE:
                return None

            tid = threading.get_ident()
            if not self._should_execute_hook(hook_type, tid):
                return None

            with ThreadSafe():
                self._maybe_update_dump_dir()
                original_state = self.ensure_gc_enabled()
                self._register_forward_hook(module, api_name)
                BaseHookManager.inner_api_count[tid] += 1
                if BaseHookManager.inner_api_count[tid] != 1:
                    return None

                full_forward_name = api_name + str(self._get_count(api_name)) + Const.SEP + Const.FORWARD
                full_backward_name = api_name + str(self._get_count(api_name)) + Const.SEP + Const.BACKWARD
                module.full_forward_name = full_forward_name
                if kwargs is None:
                    kwargs = module.msprobe_input_kwargs.get(tid, {}) if hasattr(module, 'msprobe_input_kwargs') else {}
                BaseHookManager.inner_switch[tid] = True
                module_input_output = ModuleForwardInputsOutputs(args=args, kwargs=kwargs, output=None)

                args = self._register_backward_hook(module, full_backward_name, args)
                with self._no_grad_context():
                    self.data_collector.update_api_or_module_name(full_forward_name)
                    self.data_collector.forward_input_data_collect(
                        full_forward_name,
                        module,
                        self._pid,
                        module_input_output
                    )
                BaseHookManager.inner_switch[tid] = False
                self.restore_gc_state(original_state)
                return args

        return forward_pre_hook

    def _build_forward_hook(self, hook_type, api_name):
        def forward_hook(module, args, kwargs_or_output, output_or_kwargs=None):
            tid = threading.get_ident()
            if not self._should_execute_hook(hook_type, tid):
                self._clear_input_kwargs(module, tid)
                return None

            with ThreadSafe():
                self._maybe_update_dump_dir()
                original_state = self.ensure_gc_enabled()
                if hook_type == Const.API:
                    if BaseHookManager.inner_api_count[tid] != 1:
                        if BaseHookManager.inner_api_count[tid] > 1:
                            BaseHookManager.inner_api_count[tid] -= 1
                        self._clear_input_kwargs(module, tid)
                        return None

                kwargs, output = self._process_kwargs_and_output(
                    module,
                    tid,
                    hook_type,
                    kwargs_or_output,
                    output_or_kwargs
                )
                BaseHookManager.inner_switch[tid] = True
                module_input_output = ModuleForwardInputsOutputs(args=args, kwargs=kwargs, output=output)
                if hook_type == Const.API:
                    full_forward_name = api_name + str(self._get_count(api_name)) + Const.SEP + Const.FORWARD
                    full_backward_name = api_name + str(self._get_count(api_name)) + Const.SEP + Const.BACKWARD
                    output = self._register_backward_pre_hook(module, full_backward_name, args, kwargs, output)

                with self._no_grad_context():
                    if hook_type == Const.MODULE:
                        params_dict = self._get_params_dict(module)
                        setattr(module_input_output, Const.PARAMS, params_dict)
                        if params_dict:
                            self._register_param_hook(api_name, module, params_dict)
                        self.data_collector.update_api_or_module_name(api_name)
                        self.data_collector.forward_data_collect(
                            api_name,
                            module,
                            self._pid,
                            module_input_output
                        )
                    else:
                        self.data_collector.update_api_or_module_name(full_forward_name)
                        self.data_collector.forward_output_data_collect(
                            full_forward_name,
                            module,
                            self._pid,
                            module_input_output
                        )
                        self._add_count(api_name)
                        BaseHookManager.inner_api_count[tid] -= 1
                    self._clear_input_kwargs(module, tid)
                BaseHookManager.inner_switch[tid] = False
                self.restore_gc_state(original_state)
                return output

        return forward_hook

    def _build_backward_hook(self, hook_type, full_name):
        def backward_hook(module, grad_input, grad_output):
            tid = threading.get_ident()
            if not self._should_execute_hook(hook_type, tid, is_forward=False):
                return

            with ThreadSafe():
                self._maybe_update_dump_dir()
                original_state = self.ensure_gc_enabled()
                BaseHookManager.inner_switch[tid] = True
                self.data_collector.update_api_or_module_name(full_name)

                need_exchange = self._need_exchange(module) if hook_type == Const.MODULE else True
                if need_exchange:
                    module_input_output = ModuleBackwardInputsOutputs(grad_input=grad_output, grad_output=grad_input)
                else:
                    module_input_output = ModuleBackwardInputsOutputs(grad_input=grad_input, grad_output=grad_output)
                self.data_collector.backward_data_collect(
                    full_name,
                    module,
                    self._pid,
                    module_input_output
                )
                BaseHookManager.inner_switch[tid] = False
                self.restore_gc_state(original_state)

        return backward_hook



# This file is part of the MindStudio project.
# Copyright (c) 2026 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# ==============================================================================

import os
import sqlite3

from pathlib import Path
from tensorboard.backend import http_util
from werkzeug import wrappers, Response, exceptions

from ..common.utils import Utils
from ..services.monvis_service import MonvisService


class MonvisController:
    def __init__(self, db_path):
        self.db_path = db_path
        self.monvis_service = MonvisService(self.db_path)
        self.is_db_connected = self.monvis_service.is_db_connected

    @staticmethod
    @wrappers.Request.application
    def static_file_route(request):
        filename = os.path.basename(request.path)
        extension = os.path.splitext(filename)[1]
        if extension == ".html":
            mimetype = "text/html"
        elif extension == ".js":
            mimetype = "application/javascript"
        else:
            mimetype = "application/octet-stream"

        try:
            # 添加白名单校验
            if filename != "index.html" and filename != "index.js":
                raise exceptions.NotFound("404 Not Found") from e
            server_dir = Path(__file__).resolve().parent.parent
            filepath = server_dir / "static" / filename
            with open(filepath, "rb") as infile:
                contents = infile.read()
        except IOError as e:
            raise exceptions.NotFound("404 Not Found") from e
        return Response(contents, content_type=mimetype, headers={"X-Content-Type-Options": "nosniff"})

    @wrappers.Request.application
    def request_metrics(self, request):
        """Return all available metrics and fixed stats."""

        try:
            result = self.monvis_service.get_metrics_stat()
        except sqlite3.Error as e:
            result = {"success": False, "error": f"sqlite error: {str(e)}"}
        except Exception as e:
            result = {"success": False, "error": str(e)}
        return http_util.Respond(request, result, "application/json")

    @wrappers.Request.application
    def request_values(self, request):
        """Return list of values for specified metric, stat and dimension."""
        try:
            data = Utils.safe_json_loads(request.get_data().decode("utf-8"), {})
            metric = data.get("metric")
            stat = data.get("stat")
            dimension = data.get("dimension")
            tags = data.get("tags")
            result = self.monvis_service.get_values(metric, stat, dimension, tags)
        except Exception as e:
            result = {"success": False, "error": str(e)}
        return http_util.Respond(request, result, "application/json")

    @wrappers.Request.application
    def request_tags(self, request):
        """Return list of tags for specified metric, stat and dimension."""
        try:
            metric = request.args.get("metric")
            result = self.monvis_service.get_tags(metric)
        except Exception as e:
            result = {"success": False, "error": str(e)}
        return http_util.Respond(request, result, "application/json")

    @wrappers.Request.application
    def request_heatmap_data(self, request):
        """Return heatmap data for specified parameters."""
        try:
            data = Utils.safe_json_loads(request.get_data().decode("utf-8"), {})
            metric = data.get("metric")
            stat = data.get("stat")
            dimension = data.get("dimension")
            value = data.get("value")
            tags = data.get("tags")
            result = self.monvis_service.get_heatmap_data(metric, stat, dimension, value, tags)
        except Exception as e:
            result = {"success": False, "error": str(e)}
        return http_util.Respond(request, result, "application/json")

    @wrappers.Request.application
    def request_trend_data(self, request):
        """Return trend data for specified parameters."""
        try:
            data = Utils.safe_json_loads(request.get_data().decode("utf-8"), {})
            metric = data.get("metric")
            stat = data.get("stat")
            dimension = data.get("dimension")
            dim_x = data.get("dimX")
            dim_y = data.get("dimYIdx")
            tags = data.get("tags")
            result = self.monvis_service.get_trend_data(metric, stat, dimension, dim_x, dim_y, tags)
        except Exception as e:
            result = {"success": False, "error": str(e)}
        return http_util.Respond(request, result, "application/json")



# This file is part of the MindStudio project.
# Copyright (c) 2026 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# ==============================================================================

import re
from typing import Dict, List, Tuple, Optional
from ..database.db_connection import DBConnection


class MonvisRepo:
    def __init__(self, log_dir):
        self.db = DBConnection(log_dir)
        self.conn = self.db.conn
        self.is_db_connected = self.db.is_connected()

    def _get_module_name(self, row: Dict) -> str:
        """Generate module name from row data."""
        return "_".join((str(row["target_id"]), str(row["vpp_stage"]), row["target_name"], str(row["micro_step"])))

    def _check_stat(self, stat):
        valid_stats = [
            "norm",
            "min",
            "max",
            "zeros",
            "nans",
            "mean",
            "entropy",
            "softmax_max",
            "sr",
            "kernel_norm",
            "std_x",
            "jacobian",
            "proxy",
            "token_similarity",
        ]
        if stat not in valid_stats:
            raise ValueError(f"Invalid stat: {stat}")

    def get_metric_id(self, metric_name: str) -> Optional[int]:
        """Get metric ID for a given metric name."""
        query = "SELECT metric_id FROM monitoring_metrics WHERE metric_name = ?"
        cursor = self.conn.execute(query, (metric_name,))
        row = cursor.fetchone()
        return row["metric_id"] if row else None

    def query_metrics_stat(self) -> List[str]:
        """Get all available metrics from database."""
        query = """
            SELECT * FROM global_stats 
            ORDER BY ROWID DESC 
            LIMIT 1
        """
        with self.conn as c:
            cursor = c.execute(query)
            rows = cursor.fetchone()
        return rows

    def query_global_stats(self) -> Dict[str, int]:
        """Get global statistics from database."""
        query = """
        SELECT max_rank, min_step, max_step FROM global_stats 
        ORDER BY ROWID DESC 
        LIMIT 1
        """
        cursor = self.conn.execute(query)
        row = cursor.fetchone()
        if row:
            return {"max_rank": row["max_rank"], "min_step": row["min_step"], "max_step": row["max_step"]}
        else:
            return {}

    def query_module_names_with_tags(self, tags: List[str] = None, metric_name: str = None) -> Dict[int, str]:
        """Get all module names with their target IDs, filtered by tags and metric."""
        # 检查是否存在tag_target_mapping表

        query = """
            SELECT DISTINCT mt.target_id, mt.vpp_stage, mt.target_name, mt.micro_step
            FROM monitoring_targets mt
        """

        conditions = []
        params = []
        # 添加metric过滤条件
        if metric_name:
            metric_id = self.get_metric_id(metric_name)
            if metric_id:
                # 使用trend_data表进行过滤
                conditions.append(
                    """
                    mt.target_id IN (
                        SELECT DISTINCT target_id 
                        FROM trend_data 
                        WHERE metric_id = ?
                    )
                """
                )
                params.append(metric_id)

        # 添加tag过滤条件（使用tag-target映射）
        cursor = self.conn.execute(
            """
            SELECT name FROM sqlite_master 
            WHERE type='table' AND name='tag_target_mapping'
        """
        )
        has_tag_mapping = cursor.fetchone() is not None
        if has_tag_mapping and tags:
            tag_conditions = []
            for tag in tags:
                tag_conditions.append(
                    """
                    mt.target_id IN (
                        SELECT ttm.target_id 
                        FROM tag_target_mapping ttm
                        JOIN monitoring_tags mtags ON ttm.tag_id = mtags.tag_id
                        WHERE mtags.tag_name = ?
                    )
                """
                )
                params.extend(
                    [
                        f"{tag}",
                    ]
                )
            conditions.append(f"({' AND '.join(tag_conditions)})")

        # 构建完整查询
        if conditions:
            query += " WHERE " + " AND ".join(conditions)
        cursor = self.conn.execute(query, params)
        return {row["target_id"]: self._get_module_name(row) for row in cursor}

    def query_tags(self, metric_name: str = None) -> List[Dict[str, str]]:
        """Get all available tags from database, with categories if available."""
        # 检查是否存在monitoring_tags表
        cursor = self.conn.execute(
            """
            SELECT name FROM sqlite_master 
            WHERE type='table' AND name='monitoring_tags'
        """
        )

        if cursor.fetchone():
            # 使用tags表
            query = "SELECT DISTINCT tag_name, category FROM monitoring_tags"
            params = ()

            if metric_name:
                metric_id = self.get_metric_id(metric_name)
                if metric_id:
                    # 检查是否存在tag_target_mapping表
                    cursor = self.conn.execute(
                        """
                        SELECT name FROM sqlite_master 
                        WHERE type='table' AND name='tag_target_mapping'
                    """
                    )

                    has_tag_mapping = cursor.fetchone() is not None

                    if has_tag_mapping:
                        # 使用tag-target映射表进行高效查询
                        query = """
                            SELECT DISTINCT mt.tag_name, mt.category 
                            FROM monitoring_tags mt
                            JOIN tag_target_mapping ttm ON mt.tag_id = ttm.tag_id
                            JOIN trend_data td ON ttm.target_id = td.target_id
                            WHERE td.metric_id = ?
                        """
                        params = (metric_id,)
                    else:
                        # 回退到原来的查询
                        query = """
                            SELECT DISTINCT mt.tag_name, mt.category 
                            FROM monitoring_tags mt
                            WHERE mt.metric_id = ?
                        """
                        params = (metric_id,)

            cursor = self.conn.execute(query, params)
            tags = [{"id": row["tag_name"], "text": row["tag_name"], "category": row["category"]} for row in cursor]

            # 按类别排序
            tags.sort(key=lambda x: x["category"])
            return tags
        return []

    def query_heatmap_data_with_tags(
        self, stat: str, condition: str, params: tuple, metric_id: int, tags: List[str] = None
    ) -> List[Dict]:
        """Get data for heatmap visualization with tag filtering."""
        # 检查是否存在tag_target_mapping表
        cursor = self.conn.execute(
            """
            SELECT name FROM sqlite_master 
            WHERE type='table' AND name='tag_target_mapping'
        """
        )

        has_tag_mapping = cursor.fetchone() is not None

        self._check_stat(stat)

        base_query = f"""
            SELECT t.rank, t.step, t.target_id, t.{stat}, m.target_name, m.vpp_stage, m.micro_step
            FROM trend_data t
            JOIN monitoring_targets m ON t.target_id = m.target_id
            WHERE t.metric_id = ? AND {condition}
        """
        params = (metric_id,) + params

        if has_tag_mapping and tags:
            # 使用tag-target映射表进行高效查询
            tag_conditions = []
            tag_params = []
            for tag in tags:
                tag_conditions.append(
                    """
                    t.target_id IN (
                        SELECT ttm.target_id 
                        FROM tag_target_mapping ttm
                        JOIN monitoring_tags mtags ON ttm.tag_id = mtags.tag_id
                        WHERE mtags.tag_name = ?
                    )
                """
                )
                tag_params.extend(
                    [
                        f"{tag}",
                    ]
                )

            base_query += f" AND ({' AND '.join(tag_conditions)})"
            params = params + tuple(tag_params)

        cursor = self.conn.execute(base_query, params)
        return [dict(row) for row in cursor]

    def query_trend_data_with_tags(
        self, stat: str, condition: str, params: tuple, metric_id: int, tags: List[str] = None
    ) -> List[Dict]:
        """Get data for trend visualization with tag filtering."""
        self._check_stat(stat)

        base_query = f"""
            SELECT t.step, t.rank, t.target_id, t.{stat}, m.target_name, m.vpp_stage, m.micro_step
            FROM trend_data t
            JOIN monitoring_targets m ON t.target_id = m.target_id
            WHERE t.metric_id = ? AND {condition}
        """

        # 检查是否存在tag_target_mapping表
        cursor = self.conn.execute(
            """
            SELECT name FROM sqlite_master 
            WHERE type='table' AND name='tag_target_mapping'
        """
        )
        has_tag_mapping = cursor.fetchone() is not None
        if has_tag_mapping and tags:
            # 添加tag过滤条件
            tag_conditions = []
            tag_params = []
            for tag in tags:
                tag_conditions.append(
                    """
                    t.target_id IN (
                        SELECT ttm.target_id 
                        FROM tag_target_mapping ttm
                        JOIN monitoring_tags mtags ON ttm.tag_id = mtags.tag_id
                        WHERE mtags.tag_name = ?
                    )
                """
                )
                tag_params.extend(
                    [
                        f"{tag}",
                    ]
                )

            base_query += f" AND ({' AND '.join(tag_conditions)})"
            params = params + tuple(tag_params)
        params = (metric_id,) + params

        cursor = self.conn.execute(base_query, params)
        return [dict(row) for row in cursor]
